/*
    $Id: dimm.h 98 2010-03-01 12:55:46Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _DIMM_H
#define _DIMM_H

#include <vector>

#include <boost/bind.hpp>
#include <boost/function.hpp>
#include <boost/iterator.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/vector.hpp>

namespace dimm {
	using namespace boost::numeric::ublas;

	typedef double             value_type;
	typedef vector<value_type> vector_type;
	typedef vector_type::size_type size_type;

	typedef vector_type dimm_flux_type;
	typedef vector_type dimm_vars_type;

	dimm_vars_type correction( const dimm_vars_type& vars );
	value_type turbulence( value_type scale, value_type Knumber, value_type variance );
	value_type seeing( value_type turbulence_power );
};

class dimm_storage {
	public:
		typedef boost::posix_time::ptime time_type;
		typedef dimm::vector_type vector_type;
		typedef dimm::size_type size_type;

		typedef dimm::dimm_flux_type dimm_flux_type;
		typedef dimm::dimm_vars_type dimm_vars_type;

		struct dimm_sample {
			time_type timestamp_;
			unsigned int framenum_;
			dimm_flux_type left_flux_;
			dimm_flux_type right_flux_;
			dimm_vars_type xvars_;
			dimm_vars_type yvars_;

			inline const time_type& timestamp() const { return timestamp_; }
			inline const unsigned int framenum() const { return framenum_; }
			inline const dimm_flux_type& left_flux() const { return left_flux_; }
			inline const dimm_flux_type& right_flux() const { return right_flux_; }
			inline const dimm_vars_type& xvars() const { return xvars_; }
			inline const dimm_vars_type& yvars() const { return yvars_; }

			dimm_sample( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ):
				timestamp_(time), framenum_(framenum), left_flux_(left), right_flux_(right), xvars_(xvars), yvars_(yvars) {}
		};

		typedef dimm_sample sample_type;
		typedef std::vector< sample_type > samples_list_type;

		typedef samples_list_type::const_iterator const_iterator;

		typedef boost::transform_iterator< boost::function1< dimm_flux_type, dimm_sample >, const_iterator> flux_const_iterator;
		typedef boost::transform_iterator< boost::function1< dimm_vars_type, dimm_sample >, const_iterator> vars_const_iterator;
	private:
		samples_list_type m_storage;
	public:
		dimm_storage();
		dimm_storage( const dimm_storage& storage );
		dimm_storage& operator=( const dimm_storage& storage );
		void swap( dimm_storage& storage ) throw();
		~dimm_storage();

		void accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars );
		void correct_low_vars();

		const_iterator begin() const { return m_storage.begin(); }
		const_iterator end() const { return m_storage.end(); }

		flux_const_iterator left_flux_begin() const { return flux_const_iterator( m_storage.begin(), boost::bind(&dimm_sample::left_flux,_1) ); }
		flux_const_iterator left_flux_end() const { return flux_const_iterator( m_storage.end(), boost::bind(&dimm_sample::left_flux,_1) ); }

		flux_const_iterator right_flux_begin() const { return flux_const_iterator( m_storage.begin(), boost::bind(&dimm_sample::right_flux,_1) ); }
		flux_const_iterator right_flux_end() const { return flux_const_iterator( m_storage.end(), boost::bind(&dimm_sample::right_flux,_1) ); }

		vars_const_iterator xvars_begin() const { return vars_const_iterator( m_storage.begin(), boost::bind(&dimm_sample::xvars,_1) ); }
		vars_const_iterator xvars_end() const { return vars_const_iterator( m_storage.end(), boost::bind(&dimm_sample::xvars,_1) ); }
		
		vars_const_iterator yvars_begin() const { return vars_const_iterator( m_storage.begin(), boost::bind(&dimm_sample::yvars,_1) ); }
		vars_const_iterator yvars_end() const { return vars_const_iterator( m_storage.end(), boost::bind(&dimm_sample::yvars,_1) ); }

		samples_list_type::size_type size() const { return m_storage.size(); }
};

#endif // _DIMM_H
