/*
    $Id: indices.cpp 80 2010-02-21 16:20:54Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include <algorithm>
#include <numeric>

#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/io/ios_state.hpp>

#include "atmos.h"
#include "indices.h"

#include "photometry.h"

indices::flux_type indices::flux_nonlinearity_correction ( const flux_type& flux, const vector_type& deadtime ) {
	return photometry::flux_nonlinearity_correction( flux, deadtime );
}
indices::flux_type indices::flux_background_correction( const flux_type& flux, const vector_type& scatter, const vector_type& background ) {
	return photometry::flux_scatter_correction( flux, scatter ) - background;
}
boost::tuple< indices::vars_type, indices::rho1_type, indices::rho2_type >
	indices::vars_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2, const vector_type& deadtime, const vector_type& npoisson ) {

	typedef matrix_vector_slice< vars_type > diagonal_type;
	typedef boost::tuple< vars_type, rho1_type, rho2_type > result_type;

	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );
	assert( rho1.size1() == flux.size() );
	assert( rho1.size2() == flux.size() );
	assert( rho2.size() == flux.size() );
	
	result_type ret(vars,rho1,rho2);

	diagonal_type variances( ret.get<0>(),
		slice(0, 1, std::min( ret.get<0>().size1(), ret.get<0>().size2() )),
		slice(0, 1, std::min( ret.get<0>().size1(), ret.get<0>().size2() )) );
	variances -= photometry::variance_nonpoisson_correction( flux, deadtime, npoisson );
	
	flux_type corr = apply_to_all( photometry::variance_nonlinearity_correction( flux, deadtime ), functor::inv< value_type >() );

	ret.get<0>() = element_prod( ret.get<0>(), outer_prod( corr, corr ) );
	ret.get<1>() = element_prod( ret.get<1>(), outer_prod( corr, corr ) );
	ret.get<2>() = element_prod( ret.get<2>(), element_prod( corr, corr ) );

	return ret;
}

boost::tuple< indices::indices_type, indices::indesis_type >
	indices::calculate( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {

	typedef matrix_vector_slice< indices_type >    indices_diagonal_type;
	typedef matrix_vector_slice< const rho1_type > rho1_diagonal_type;
	typedef boost::tuple< indices_type, indesis_type> result_type;

	assert( vars.size1() == rho1.size1() );
	assert( vars.size2() == rho1.size2() );
	assert( vars.size1() == flux.size() );
	assert( vars.size2() == flux.size() );

	result_type ret;
	
	ret.get<0>() = element_prod( vars*1.25 - rho1*0.25, outer_prod(apply_to_all(flux,functor::inv< value_type >()),apply_to_all(flux,functor::inv< value_type >())) );

	indices_diagonal_type variances( ret.get<0>(),
		slice(0, 1, std::min( ret.get<0>().size1(), ret.get<0>().size2() )),
		slice(0, 1, std::min( ret.get<0>().size1(), ret.get<0>().size2() )) );
	rho1_diagonal_type rho1_diag( rho1,
		slice(0, 1, std::min( rho1.size1(), rho1.size2() )),
		slice(0, 1, std::min( rho1.size1(), rho1.size2() )) );

	triangular_adaptor< vars_type, unit_lower >( ret.get<0>() ) = outer_prod( variances, scalar_vector< value_type >( variances.size(), 1.0 ) ) + outer_prod( scalar_vector< value_type >( variances.size(), 1.0 ), variances ) - triangular_adaptor< vars_type, unit_lower >( ret.get<0>() ) * 2.0;

	ret.get<1>() = ( variances*3.0 + element_prod( rho2 - rho1_diag * 4.0, apply_to_all( element_prod(flux,flux), functor::inv< value_type >() ) ) ) / 4.5;

	return ret;
}

indices_factory::indices_factory() {
	m_samples_list.reserve( 60 );
}
indices_factory::indices_factory( const indices_factory& factory ) {
	typedef filter_list_type::const_iterator filter_iterator;

	for( filter_iterator it = factory.m_filter_list.begin(); it != factory.m_filter_list.end(); ++it ) {
		m_filter_list.push_back( (*it)->clone() );
	}

	m_samples_list = factory.m_samples_list;
}
void indices_factory::swap(indices_factory& factory) throw() {
	factory.m_filter_list.swap( m_filter_list );
	factory.m_samples_list.swap( m_samples_list );
}
indices_factory& indices_factory::operator=( const indices_factory& factory ) {
	indices_factory tmp( factory );
	tmp.swap( *this );
	return *this;
}
indices_factory::~indices_factory() {
	typedef filter_list_type::iterator filter_iterator;

	for( filter_iterator it = m_filter_list.begin(); it != m_filter_list.end(); ++it ) {
		delete (*it);
	}
}

bool indices_factory::flux_filter( const time_type& time, const flux_type& flux, bool wexcept ) const {
	typedef filter_list_type::const_iterator const_iterator;

	bool filter = true;
	for( const_iterator it = m_filter_list.begin(); it != m_filter_list.end(); ++it ) {
		filter &= (*it)->flux_filter( time, flux, wexcept ); // can throw here
	}
	return filter;
}
void indices_factory::accumulate( const time_type& time, const flux_type& flux, const indices_type& indices, const indesis_type& indesis ) {
	m_samples_list.push_back( boost::make_tuple( time, flux, indices, indesis ) );
}

indices_abstract_filter::indices_abstract_filter() {}
indices_abstract_filter::~indices_abstract_filter() {}

indices_threshold_filter::indices_threshold_filter( value_type threshold ):
	m_threshold( threshold ) {
}
indices_threshold_filter::~indices_threshold_filter() {
}
bool indices_threshold_filter::flux_filter( const time_type& time, const flux_type& flux, bool wexcept ) const {
	if( *std::min_element( flux.begin(), flux.end() ) < m_threshold ) {
		if( wexcept )
			throw std::logic_error( (boost::format("Object flux %.2f is under the threshold %.2f") % flux % m_threshold).str() );
		return false;
	}
	return true;
}
indices_abstract_filter* indices_threshold_filter::clone() const {
	return new indices_threshold_filter( *this );
}
